from abc import ABCMeta, abstractmethod


class Estimator(metaclass=ABCMeta):
    """Abstract base class for estimator.

    An estimator is the master class of all modeling operations. At minimum,
    it includes:

    1. input data and output data transformations (e.g. input image cropping,
    remove unused output labels...) when applicable.
    2. neural network graph (model) for either pytorch or tensorflow.
    3. procedures to execute model training and evaluation.

    One estimator could support multiple tasks (e.g. Mask R-CNN can be used for
    semantic segmentation and object detection)
    """

    @staticmethod
    def create(name, **kwargs):
        """Create a new instance of the estimators subclass

        Args:
            name (str): unique identifier for a estimators subclass
            config (dict): parameters specific to each estimators subclass
                used to create a estimators instance

        Returns:
            an instance of the specified estimators subclass
        """
        estimators_cls = Estimator.find(name)
        return estimators_cls(**kwargs)

    @staticmethod
    def find(name):
        """Find Estimator subclass based on the given name

        Args:
            name (str): unique identifier for a estimators subclass

        Returns:
            a label of the specified estimators subclass
        """
        estimators_classes = Estimator.__subclasses__()
        estimators_names = [e.__name__ for e in estimators_classes]
        if name in estimators_names:
            estimators_cls = estimators_classes[estimators_names.index(name)]
            return estimators_cls
        else:
            raise NotImplementedError(f"Unknown Estimator class {name}!")

    @abstractmethod
    def train(self, **kwargs):
        """Abstract method to train estimators
        """
        raise NotImplementedError("Subclass needs to implement this method")

    @abstractmethod
    def evaluate(self, **kwargs):
        """Abstract method to evaluate estimators
        """
        raise NotImplementedError("Subclass needs to implement this method")



import logging
from pathlib import Path

import datasetinsights.constants as const
from datasetinsights.data.simulation import Metrics, MetricDefinitions
from datasetinsights.data.simulation.download import Downloader
from datasetinsights.data.simulation.tables import SCHEMA_VERSION


logger = logging.getLogger(__name__)


def download(data_root, manifest_file, use_cache=True):
    """Download dataset for dataset statistics
    Args:
        root (str): root directory where the dataset should be downloaded.
        manifest_file (str): path to USim simulation manifest file.
        use_cache (bool): set to False to override already downloaded files

    Returns
        str: root directory where the dataset is downloaded.
    """
    path = Path(data_root)
    path.mkdir(parents=True, exist_ok=True)

    dl = Downloader(manifest_file, data_root, use_cache=use_cache)
    dl.download_metrics()
    dl.download_references()
    dl.download_captures()

    return data_root


class RenderedObjectInfo:
    """Rendered Object Info in Captures

    This metric stores common object info captured by a sensor in the simulation
    environment. It can be used to calculate object statistics such as
    object count, object rotation and visible pixels.

    Attributes:
        raw_table (pd.DataFrame): rendered object info stored with a tidy
        pandas dataframe. Columns "label_id", "instance_id", "visible_pixels",
        "capture_id, "label_name".

    Examples:
        >>> # set the data root path to where data was stored
        >>> data_root = "$HOME/data"
        >>> # use rendered object info definition id
        >>> definition_id = "659c6e36-f9f8-4dd6-9651-4a80e51eabc4"
        >>> roinfo = RenderedObjectInfo(data_root, definition_id)

        total object count per label dataframe
        >>> roinfo.total_counts()
        label_id label_name count
               1    object1    10
               2    object2    21

        object count per capture dataframe
        >>> roinfo.per_capture_counts()
        capture_id  count
            qwerty     10
            asdfgh     21
    """
    LABEL = "label_id"
    LABEL_READABLE = "label_name"
    INDEX_COLUMN = "capture_id"
    VALUE_COLUMN = "values"
    COUNT_COLUMN = "count"

    def __init__(
        self,
        data_root=const.DEFAULT_DATA_ROOT,
        version=SCHEMA_VERSION,
        def_id=None,
    ):
        """Initialize RenderedObjectInfo

        Args:
            data_root (str): root directory where the dataset was stored
            version (str): synthetic dataset schema version
            def_id (str): rendered object info definition id
        """
        filtered_metrics = Metrics(data_root, version).filter_metrics(def_id)
        label_mappings = self._read_label_mappings(data_root, version, def_id)
        self.raw_table = self._read_filtered_metrics(
            filtered_metrics, label_mappings
        )

    def num_captures(self):
        """Total number of captures

        Returns:
            integer: Total number of captures
        """
        return self.raw_table[self.INDEX_COLUMN].nunique()

    @staticmethod
    def _read_label_mappings(data_root, version, def_id):
        """Read label_mappings from a metric_definition record.

        Args:
            data_root (str): root directory where the dataset was stored
            version (str): synthetic dataset schema version
            def_id (str): rendered object info definition id

        Returns:
            dict: The mappings of {label_id: label_name}
        """
        definition = (
            MetricDefinitions(data_root, version).get_definition(def_id)
        )
        name = RenderedObjectInfo.LABEL
        readable_name = RenderedObjectInfo.LABEL_READABLE

        return {d[name]: d[readable_name] for d in definition["spec"]}

    @staticmethod
    def _read_filtered_metrics(filtered_metrics, label_mappings):
        """Read label_mappings from a metric_definition record.

        Args:
            filtered_metrics (pd.DataFrame): A pandas dataframe for metrics
                filtered by definition id.
            label_mappings (dict): the mappings of {label_id: label_name}

        Returns:
            pd.DataFrame: rendered object info stored with a tidy
            pandas dataframe. Columns "label_id", "instance_id",
            "visible_pixels", "capture_id, "label_name".
        """
        filtered_metrics[RenderedObjectInfo.LABEL_READABLE] = (
            filtered_metrics[RenderedObjectInfo.LABEL].map(label_mappings)
        )
        # Remove metrics data not defined in label_mappings
        filtered_metrics.dropna(
            subset=[RenderedObjectInfo.LABEL_READABLE], inplace=True
        )

        return filtered_metrics

    def total_counts(self):
        """Aggregate Total Object Counts Per Label

        Returns:
            pd.DataFrame: Total object counts table.
                Columns "label_id", "label_name", "count"
        """
        agg = (
            self.raw_table
            .groupby([self.LABEL, self.LABEL_READABLE])
            .size()
            .to_frame(name=self.COUNT_COLUMN)
            .reset_index()
        )

        return agg

    def per_capture_counts(self):
        """ Aggregate Object Counts Per Label

        Returns:
            pd.DataFrame: Total object counts table.
                Columns "capture_id", "count"
        """
        agg = (
            self.raw_table
            .groupby(self.INDEX_COLUMN)
            .size()
            .to_frame(name=self.COUNT_COLUMN)
            .reset_index()
        )

        return agg



from pathlib import Path
import json
import logging
import os
import zipfile

from PIL import Image

import datasetinsights.constants as const
from datasetinsights.data.bbox import BBox2D
from datasetinsights.storage.gcs import GCSClient
from .base import Dataset


GROCERIES_GCS_PATH = "data/groceries"
GROCERIES_LOCAL_PATH = "groceries"
JSON_NAME = "annotations"
DATA_PREFIX_NAME = "groceries_real"
ANNOTATION_FILE_NAME = "annotation_definitions"
SPLITS = ("train", "val", "test")
VERSIONS = ("v2", "v3")

logger = logging.getLogger(__name__)


class GroceriesReal(Dataset):
    """Groceries Real dataset loader.

    Attributes:
        root (str): local directory where groceries_real dataset is saved
        groceries_real_data (list): training or testing data
        split (str): indicate split type of the dataset (train|val|test)
        transforms: callable transformation
        label_mappings (dict): a dict of {label_id: label_name} mapping
        version (str): version of GroceriesReal dataset, e.g. "v2", "v3".
        default version="v3".
    """

    def __init__(
        self,
        *,
        data_root=const.DEFAULT_DATA_ROOT,
        split="train",
        transforms=None,
        version="v3",
        **kwargs,
    ):
        """
        Args:
            data_root (str): root directory prdfix of all datasets
            split (str): indicate split type of the dataset (train|val|test)
            transforms: callable transformation
            version (str): version of GroceriesReal dataset
            groceries_data (list): a list of all image info in the json file
            data_indices (list): a list of indices for image info in the json
            file
            In annotations.json file, all the image info are saved in a list:
            [
                {
                "file_name": "IMG_4185.JPG",
                "annotations":[{
                    "label_id": 11,
                    "bboxs":[1692, 1386, 768, 133
                }]
                },
            ]
            In the txt files, it contains all the indices for image info in the
            json file:
            for example, in grocieres_real_train.txt, it looks like:
                0 1 2 3 4 5 6 7 8 9 10 11 12 13
                which represents indices of train data in annotations.json file
        """
        self.split = split
        if split not in SPLITS:
            raise ValueError(
                f"invalid value for split: {split}, possible values "
                f"are: {SPLITS}"
            )
        logger.info(f"split for groceries real is {split}")
        if version not in VERSIONS:
            raise ValueError(
                f"A valid dataset version should be set. "
                f"Available versons: {VERSIONS}"
            )
        self.root = os.path.join(data_root, GROCERIES_LOCAL_PATH)
        self.version = version
        self.download()
        self.groceries_data = self._get_annotations()
        self.data_indices = self._get_data_indices()
        self.transforms = transforms
        self.label_mappings = self._get_label_mappings()

    def __getitem__(self, idx):
        """
        Args:
            idx (int): index

        Returns:
            A tuple of image name and a list of bounding boxes inside the image
        """
        data_idx = self.data_indices[idx]
        image_data = self.groceries_data[data_idx]
        image_name = image_data["file_name"]
        annotations = image_data["annotations"]
        image = Image.open(
            os.path.join(self.root, self.version, "images", image_name)
        )
        bboxes = self._convert_annotations(annotations)
        if self.transforms:
            return self.transforms(image, bboxes)
        else:
            return (image, bboxes)

    def __len__(self):
        return len(self.data_indices)

    def _get_local_data_zip(self):
        """create a local path for download zip file
        """
        return os.path.join(self.root, f"{self.version}.zip")

    def _get_annotations(self):
        """Read data from the annotations.json.
        In the json file, image file name and annotations are saved in a list.
        For example:
            [
                {
                "file_name": "IMG_4185.JPG",
                "annotations":[{
                    "label_id": 11,
                    "bboxs":[1692, 1386, 768, 133
                }]
                },
            ]

        Return:
            json_data (list): a list of data in annotations.json file
        """
        json_data = json.load(
            open(os.path.join(self.root, self.version, f"{JSON_NAME}.json"))
        )

        return json_data

    def _convert_annotations(self, annotations):
        """Convert the bounding boxes for an image into canonical 2d bbox.

        Args:
            annotations (list): bounding box information from annotations.json

        Return:
            bboxes (list): a list of bounding box for an image
        """
        bboxes = []
        for ann in annotations:
            bbox = self._convert_canonical_2Dbbox(ann)
            bboxes.append(bbox)

        return bboxes

    def download(self):
        """download dataset from GCS
        """
        path = Path(self.root)
        path.mkdir(parents=True, exist_ok=True)
        client = GCSClient()
        data_zip_gcs = os.path.join(GROCERIES_GCS_PATH, f"{self.version}.zip")

        data_zip_local = self._get_local_data_zip()
        if not os.path.exists(data_zip_local):
            logger.info(f"no data zip file found, will download.")
            client.download(
                bucket_name=const.GCS_BUCKET,
                object_key=data_zip_gcs,
                localfile=data_zip_local,
            )

            with zipfile.ZipFile(data_zip_local, "r") as zip_dir:
                zip_dir.extractall(f"{self.root}")

    def _get_data_indices(self):
        """Get data indices from txt file.

        Returns:
            idx_data (list): a list of data indices in self.groceries_data
        """
        file_path = os.path.join(
            self.root, self.version, f"{DATA_PREFIX_NAME}_{self.split}.txt"
        )
        with open(file_path) as txt_file:
            idx_data = [int(i) for i in txt_file.readline().split()]

        return idx_data

    def _get_label_mappings(self):
        """Get data indices from txt file.

        Returns:
            label_mappings (dict): a dict of {label_id: label_name} mapping
        """
        file_path = os.path.join(
            self.root, self.version, f"{ANNOTATION_FILE_NAME}.json"
        )
        with open(file_path) as json_file:
            init_definition = json.load(json_file)["annotation_definitions"][0]
            label_mappings = {
                m["label_id"]: m["label_name"] for m in init_definition["spec"]
            }

        return label_mappings

    def _convert_canonical_2Dbbox(self, single_bbox):
        """Convert BBobx to canonical BBox2D.

        Args:
            single_bbox (dict): raw bounding box information

        Return:
            canonical_bbox (BBox2D): canonical bounding box
        """
        label = single_bbox["label_id"]
        bbox = single_bbox["bbox"]

        canonical_bbox = BBox2D(
            x=bbox[0], y=bbox[1], w=bbox[2], h=bbox[3], label=label
        )
        return canonical_bbox



import copy
import logging
import math
from typing import Dict, List, Tuple

from codetiming import Timer
import numpy as np
import torch
import torch.distributed as dist
import torchvision

import datasetinsights.constants as const
from datasetinsights.data.bbox import BBox2D
from datasetinsights.data.datasets import Dataset
from datasetinsights.data.transforms import Compose
from datasetinsights.evaluation_metrics.base import EvaluationMetric
from datasetinsights.torch_distributed import get_world_size
from .base import Estimator


MAX_BOXES_PER_IMAGE = 100
logger = logging.getLogger(__name__)
DEFAULT_ACCUMULATION_STEPS = 1


class BadLoss(Exception):
    pass


class Loss:
    """
    Record Loss during epoch
    """

    def __init__(self):
        self._sum = 0
        self.num_examples = 0

    def reset(self):
        self._sum = 0
        self.num_examples = 0

    def update(self, avg_loss, batch_size):
        self._sum += avg_loss * batch_size
        self.num_examples += batch_size

    def compute(self):
        """

        Returns (float): avg. loss

        """
        if self.num_examples == 0:
            raise ValueError(
                "Loss must have at least one example "
                "before it can be computed."
            )
        return self._sum / self.num_examples


def convert_bboxes2canonical(bboxes):
    """
    convert bounding boxes from the format used by pytorch torchvision's
    faster rcnn model into our canonical format, a list of list of BBox2Ds.
    Faster RCNN format:
    https://github.com/pytorch/vision/blob/master/torchvision/models/
    detection/faster_rcnn.py#L45

    Args:
        bboxes (List[Dict[str, torch.Tensor()): A list of dictionaries. Each
        item in the list corresponds to the bounding boxes for one example.
        The dictionary must have the keys 'boxes' and 'labels'. The value for
        'boxes' is (``FloatTensor[N, 4]``): the ground-truth boxes in
        ``[x1, y1, x2, y2]`` format, with values between ``0`` and ``H`` and
        ``0`` and ``W``. The value for labels is (``Int64Tensor[N]``): the
        class label for each ground-truth box. If the dictionary has the key
        `scores` then these values are used for the confidence score of the
        BBox2D, otherwise the score is set to 1.

    Returns (list[List[BBox2D]]): Each element in the list corresponds to the
    list of bounding boxes for an example.

    """
    bboxes_batch = []
    for example in bboxes:
        all_coords = example["boxes"]
        num_boxes = all_coords.shape[0]
        labels = example["labels"]
        if "scores" in example.keys():
            scores = example["scores"]
        else:
            scores = torch.FloatTensor([1.0] * num_boxes)
        bboxes_example = []
        for i in range(num_boxes):
            coords = all_coords[i]
            x, y = coords[0].item(), coords[1].item()
            canonical_box = BBox2D(
                x=x,
                y=y,
                w=coords[2].item() - x,
                h=coords[3].item() - y,
                label=labels[i].item(),
                score=scores[i].item(),
            )
            bboxes_example.append(canonical_box)
        bboxes_batch.append(bboxes_example)
    return bboxes_batch


def reduce_dict(input_dict, average=True):
    """
    Args:
        input_dict (dict): all the values will be reduced
        average (bool): whether to do average or sum
    Reduce the values in the dictionary from all processes so that all processes
    have the averaged results. Returns a dict with the same fields as
    input_dict, after reduction.
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict
    with torch.no_grad():
        names = []
        values = []
        # sort the keys so that they are consistent across processes
        for k in sorted(input_dict.keys()):
            names.append(k)
            values.append(input_dict[k])
        values = torch.stack(values, dim=0)
        dist.all_reduce(values)
        if average:
            values /= world_size
        reduced_dict = {k: v for k, v in zip(names, values)}
    return reduced_dict


def prepare_bboxes(bboxes: List[BBox2D]) -> Dict[str, torch.Tensor]:
    """Prepare bounding boxes for model training.

    Args:
        bboxes: mini batch of bounding boxes (not including images).
        Each example is a list of bounding boxes. Torchvision's implementation
        of Faster-RCNN requires bounding boxes to be in the format of a
        dictionary {'labels':[label ids, ...], and
        'boxes': [[xleft, ytop, xright, ybottom of box1],
        [xleft, ytop, xright, ybottom of box2]...]

    Returns: bounding boxes in the form that Faster RCNN expects
    """
    labels = []
    box_list = []
    for b in bboxes:
        labels.append(b.label)
        box = [b.x, b.y, b.x + b.w, b.y + b.h]
        if box[0] >= box[2] or box[1] >= box[3]:
            raise ValueError(
                f"box not properly formed with coordinates: "
                f"{box} [xleft, ytop, xright, ybottom]"
            )
        box_list.append(box)

    d = {"boxes": torch.Tensor(box_list), "labels": torch.LongTensor(labels)}
    return d


def train_one_epoch(
    *,
    writer,
    model,
    optimizer,
    data_loader,
    device,
    epoch,
    lr_scheduler,
    accumulation_steps,
    log_frequency=1000,
):
    """

    Args:
        writer: Tensorboard writer object
        model: pytorch model
        optimizer: pytorch optimizer
        data_loader(DataLoader): pytorch dataloader
        device: model training on device (cpu|cuda)
        epoch (int):
        accumulation_steps (int): Accumulated Gradients are only updated after
        X steps. This creates an effective batch size of batch_size *
        accumulation_steps
        lr_scheduler: Pytorch LR scheduler
        log_frequency: after this number of mini batches log average loss
    """
    model.train()
    loss_metric = Loss()
    optimizer.zero_grad()
    n_examples = len(data_loader.dataset)
    for i, (images, targets) in enumerate(data_loader):
        images = list(image.to(device) for image in images)
        targets = [{k: v.to(device) for k, v in t.items()} for t in targets]
        loss_dict = model(images, targets)

        losses_grad = sum(loss for loss in loss_dict.values())

        loss_dict_reduced = reduce_dict(loss_dict)
        losses = sum(loss for loss in loss_dict_reduced.values())
        loss_metric.update(avg_loss=losses.item(), batch_size=len(targets))

        if not math.isfinite(losses):
            raise BadLoss(
                f"Loss is {losses}, stopping training. Input was "
                f"{images, targets}. Loss dict is {loss_dict}"
            )
        elif i % log_frequency == 0:
            intermediate_loss = loss_metric.compute()
            examples_seen = epoch * n_examples + loss_metric.num_examples
            logger.info(
                f"intermediate loss after mini batch {i} in epoch {epoch} "
                f"(total training examples: {examples_seen}) is "
                f"{intermediate_loss}"
            )
            writer.add_scalar(
                "training/running_loss", intermediate_loss, examples_seen
            )
        losses_grad.backward()
        if (i + 1) % accumulation_steps == 0:
            optimizer.step()
            lr_scheduler.step()
            optimizer.zero_grad()
    writer.add_scalar("training/loss", loss_metric.compute(), epoch)
    writer.add_scalar(
        "training/running_loss", loss_metric.compute(),
        loss_metric.num_examples
    )
    writer.add_scalar("training/lr", optimizer.param_groups[0]["lr"], epoch)
    loss_metric.reset()


def canonical2list(bbox: BBox2D):
    """
    convert a BBox2d into a single list
    Args:
        bbox:

    Returns:

    """
    return [bbox.label, bbox.score, bbox.x, bbox.y, bbox.w, bbox.h]


def list2canonical(box_list):
    """
    convert a list into a Bbox2d
    Args:
        box_list: box represented in list format

    Returns (BBox2d):

    """
    return BBox2D(label=box_list[0], score=box_list[1], x=box_list[2],
                  y=box_list[3], w=box_list[4], h=box_list[5])


def list3d_2canonical(batch):
    """
    convert a list of list of padded targets and predictions per examples where
    bounding boxes are represented by lists into the same format except
    the boxes are represented by the BBox2d class and the padded boxes are
    removed
    Args:
        batch: [[[gt], [prds]], [[gt], [prds]], ] where gt and prds are list
        of lists

    Returns:
        [([gt],[preds]), ([gt],[preds])... where gt and preds are lists of
        BBox2ds
    """
    all_canonical = []
    for example in batch:
        targets, preds = example
        # todo should break after first nan
        targets_canonnical = [
            list2canonical(target_list)
            for target_list in targets
            if not np.isnan(target_list).any()
        ]
        preds = [
            list2canonical(pred_list)
            for pred_list in preds
            if not np.isnan(pred_list).any()
        ]
        all_canonical.append((targets_canonnical, preds))
    return all_canonical


def pad_box_lists(
    gt_preds: List[Tuple[List[BBox2D], List[BBox2D]]],
    max_boxes_per_img=MAX_BOXES_PER_IMAGE,
):
    """
    Pad the list of boxes and targets with place holder boxes so that all
    targets and predictions have the same number of elements
    Args:
        gt_preds (list(tuple(list(BBox2d), (Bbox2d)))): A list of tuples where
        the first element in each tuple is a list of bounding boxes
        corresponding to the targets in an example, and the second element
        in the tuple corresponds to the predictions in that example
        max_boxes_per_img: : maximum number of target boxes and predicted boxes
         per image

    Returns: same format as gt_preds but all examples will have the same
    number of targets and predictions. If there are fewer targets or
    predictions than max_boxes_per_img, then boxes with nan values are added.

    """
    padding_box = BBox2D(
        label=np.nan, score=np.nan, x=np.nan, y=np.nan, w=np.nan, h=np.nan
    )
    for tup in gt_preds:
        target_list, pred_list = tup
        if len(target_list) > max_boxes_per_img:
            raise ValueError(
                f"max boxes per image set to {max_boxes_per_img},"
                f" but there were {len(target_list)} targets"
                f" found."
            )
        if len(pred_list) > max_boxes_per_img:
            raise ValueError(
                f"max boxes per image set to {max_boxes_per_img},"
                f" but there were {len(target_list)} predictions"
                f" found."
            )
        for i in range(max_boxes_per_img - len(target_list)):
            target_list.append(padding_box)
        for i in range(max_boxes_per_img - len(pred_list)):
            pred_list.append(padding_box)
    return gt_preds


def _gt_preds2tensor(gt_preds, max_boxes=MAX_BOXES_PER_IMAGE):
    """

    Args:
        gt_preds (list(tuple(list(BBox2d), (Bbox2d)))): A list of tuples where
        the first element in each tuple is a list of bounding boxes
        corresponding to the targets in an example, and the second element
        in the tuple corresponds to the predictions in that example
        max_boxes: maximum number of target boxes and predicted boxes per image

    Returns (Tensor): [[gt, preds], [gt, preds]] where gt and preds are both
    tensors where each element is a 1-d tensor representing a bbox

    """
    gt_preds_padded = pad_box_lists(gt_preds, max_boxes)
    padded_list = []
    for i in range(len(gt_preds_padded)):
        gt_list, target_list = gt_preds_padded[i]
        gt_tensor, target_tensor = (
            [canonical2list(b) for b in gt_list],
            [canonical2list(b) for b in target_list],
        )
        gt_preds_tensor = [gt_tensor, target_tensor]
        padded_list.append(gt_preds_tensor)
    gt_preds_tensor = torch.Tensor(padded_list)
    return gt_preds_tensor


def gather_gt_preds(*, gt_preds, device, max_boxes=MAX_BOXES_PER_IMAGE):
    """

    Args:
        gt_preds (list(tuple(list(BBox2d), (Bbox2d)))): A list of tuples where
        the first element in each tuple is a list of bounding boxes
        corresponding to the targets in an example, and the second element
        in the tuple corresponds to the predictions in that example
        device:
        max_boxes: the maximum number of boxes allowed for either targets or
        predictions per image

    Returns (list(tuple(list(BBox2d), (Bbox2d)))): a list in the same format
    as gt_preds but containing all the information across processes e.g. if
    rank 0 has gt_preds = [([box_0], []), ([box_1], [box_2, box_2.5])] and
    rank 1 has gt_preds [([], [box_3]), ([box_4, box_5], [])] then this
    function will return [([box_0], []), ([box_1], [box_2, box_2.5]), ([],
    [box_3]), ([box_4, box_5], [])] the returned list is consistent across all
    processes

    """
    gt_preds_tensor = _gt_preds2tensor(gt_preds, max_boxes)

    all_preds_gt = [
        torch.empty(list(gt_preds_tensor.size())).to(device)
        for i in range(get_world_size())
    ]
    dist.all_gather(tensor_list=all_preds_gt, tensor=gt_preds_tensor.to(device))
    all_preds_gt_canonical = tensorlist2canonical(all_preds_gt)
    return all_preds_gt_canonical


def tensorlist2canonical(tensor_list):
    """
    Converts the gt and predictions into the canonical format and removes
    the boxes with nan values that were added for padding
    Args:
        tensor_list: [tensor([[gt, prds]), tensor([gt, prds])], ...]

    Returns (list(tuple(list(BBox2d), (Bbox2d)))): A list of tuples where
        the first element in each tuple is a list of bounding boxes
        corresponding to the targets in an example, and the second element
        in the tuple corresponds to the predictions in that example

    """
    gt_preds_canonical = []
    for t in tensor_list:
        gt_preds_canonical += list3d_2canonical(t.tolist())
    return gt_preds_canonical


@torch.no_grad()
def evaluate(
    *,
    model,
    data_loader,
    device,
    writer,
    epoch,
    metrics,
    label_mappings,
    max_detections_per_img=MAX_BOXES_PER_IMAGE,
    is_distributed=False,
    synchronize_metrics=True,
):
    """
    Evaluate model performance. Note, torchvision's implementation of faster
    rcnn requires input and gt data for training mode and returns a dictionary
    of losses (which we need to record the loss). We also need to get the raw
    predictions, which is only possible in model.eval() mode, to calculate
    the evaluation metric.
    Args:
        model: pytorch model
        data_loader (DataLoader): pytorch dataloader
        device: model training on device (cpu|cuda)
        writer: Tensorboard writer object
        epoch (int): current epoch, used for logging
        metrics (dict): a dict of metrics (key: metric name; value: metric)
        label_mappings (dict): a dict of {label_id: label_name} mapping
        max_detections_per_img: max number of targets or predictions allowed
        per example
        is_distributed: whether or not the model is distributed
        synchronize_metrics: whether or not to synchronize evaluation metrics
        across processes
    Returns:

    """
    n_threads = torch.get_num_threads()
    torch.set_num_threads(1)
    model.eval()

    loss_metric = Loss()
    for metric in metrics.values():
        metric.reset()
    for i, (images, targets) in enumerate(data_loader):
        images = list(image.to(device) for image in images)
        targets_raw = [{k: v.to(device) for k, v in t.items()} for t in targets]
        targets_converted = convert_bboxes2canonical(targets)
        # we need 2 do forward passes, one in eval mode gets us the raw preds
        # and one in train mode gets us the losses
        # pass number 1 to get mAP
        model.eval()
        preds_raw = model(images)
        converted_preds = convert_bboxes2canonical(preds_raw)
        gt_preds = list(zip(targets_converted, converted_preds))
        if is_distributed and synchronize_metrics:
            all_preds_gt_canonical = gather_gt_preds(
                gt_preds=gt_preds,
                device=device,
                max_boxes=max_detections_per_img,
            )
        else:
            all_preds_gt_canonical = gt_preds
        for metric in metrics.values():
            metric.update(all_preds_gt_canonical)

        # pass number 2 to get val loss
        model.train()
        loss_dict = model(images, targets_raw)
        loss_dict_reduced = reduce_dict(loss_dict)
        losses = sum(loss for loss in loss_dict_reduced.values())
        loss_metric.update(avg_loss=losses.item(), batch_size=len(targets))

    for metric_name, metric in metrics.items():
        result = metric.compute()
        logger.debug(result)
        mean_result = np.mean(
            [result_per_label for result_per_label in result.values()]
        )
        logger.debug(f"metric {metric_name} has mean result: {mean_result}")
        writer.add_scalar(f"val/m{metric_name}", mean_result, epoch)

        label_results = {label_mappings[id]: result[id] for id in result}
        writer.add_scalars(f"val/{metric_name}-per-class", label_results, epoch)

    val_loss = loss_metric.compute()
    logger.info(f"validation loss is {val_loss}")
    writer.add_scalar("val/loss", val_loss, epoch)
    torch.set_num_threads(n_threads)


class PrepBoxes:
    def __call__(self, image, target):
        target = prepare_bboxes(target)
        return image, target


class ToTensor:
    def __call__(self, image, target):
        image = torchvision.transforms.functional.to_tensor(image)
        return image, target


def get_transform():
    transforms = [PrepBoxes(), ToTensor()]
    return Compose(transforms)


def collate_fn(batch):
    """
    Prepare batch to be format Faster RCNN expects
    Args:
        batch: mini batch of the form ((x0,x1,x2...xn),(y0,y1,y2...yn))

    Returns:
        mini batch in the form [(x0,y0), (x1,y2), (x2,y2)... (xn,yn)]
    """
    return tuple(zip(*batch))


class FasterRCNN(Estimator):
    """
    https://github.com/pytorch/vision/tree/master/references/detection
    https://arxiv.org/abs/1506.01497
    Args:
        config (CfgNode): estimator config
        writer: Tensorboard writer object
        checkpointer: Model checkpointer callback to save models
        device: model training on device (cpu|cuda)
        local_rank (int): (optional) rank of process executing code
        gpu (int): (optional) gpu id on which code will execute
    Attributes:
        model: pytorch model
        writer: Tensorboard writer object
        checkpointer: Model checkpointer callback to save models
        device: model training on device (cpu|cuda)
    """

    def __init__(
        self, *, config, writer, checkpointer, device, gpu=0, rank=0, **kwargs
    ):
        logger.info(f"initializing faster rcnn")
        self.config = config
        self.device = device
        self.writer = writer
        model_name = f"fasterrcnn_{self.config.backbone}_fpn"
        self.model = torchvision.models.detection.__dict__[model_name](
            num_classes=config.num_classes,
            pretrained_backbone=config.pretrained_backbone,
            pretrained=config.pretrained,
            box_detections_per_img=MAX_BOXES_PER_IMAGE,
        )
        self.model_without_ddp = self.model
        self.gpu = gpu
        self.rank = rank
        self.sync_metrics = config.get("synchronize_metrics", True)
        logger.info(f"gpu: {self.gpu}, rank: {self.rank}")

        self.checkpointer = checkpointer
        checkpoint_file = config.get("checkpoint_file", const.NULL_STRING)
        if checkpoint_file != const.NULL_STRING:
            checkpointer.load(self, config.checkpoint_file)

        self.metrics = {}
        for metric_key, metric in config.metrics.items():
            self.metrics[metric_key] = EvaluationMetric.create(
                metric.name, **metric.args
            )
        self.model.to(self.device)

        if self.config.system.distributed:
            self.model = torch.nn.parallel.DistributedDataParallel(
                self.model, device_ids=[self.gpu]
            )
            self.model_without_ddp = self.model.module

    def evaluate(self, **kwargs):
        config = self.config
        test_dataset = Dataset.create(
            config.test.dataset.name,
            data_root=config.system.data_root,
            transforms=get_transform(),
            **config.test.dataset.args,
        )
        if self.config.system.dryrun:
            test_dataset = self._create_dryrun_dataset(
                dataset=test_dataset, subset_size=self.config.val.batch_size * 2
            )
        test_sampler = self._create_sampler(
            dataset=test_dataset, is_train=False,
        )

        logger.info("Start evaluating estimator: %s", type(self).__name__)
        test_loader = self._create_loader(
            dataset=test_dataset,
            batch_size=self.config.val.batch_size,
            sampler=test_sampler,
            collate_fn=collate_fn,
            train=False,
        )
        self.model.to(self.device)
        evaluate(
            model=self.model,
            data_loader=test_loader,
            device=self.device,
            writer=self.writer,
            epoch=0,
            metrics=self.metrics,
            label_mappings=test_dataset.label_mappings,
            is_distributed=self.config.system.distributed,
            synchronize_metrics=self.sync_metrics,
        )

    def _create_sampler(self, *, dataset, is_train):
        """

        Args:
            dataset: dataset obj must have len and __get_item__
            is_train: whether or not the sampler is for training data

        Returns:
            https://pytorch.org/docs/stable/data.html#torch.utils.data.Sampler

        """
        if self.config.system.distributed:
            sampler = torch.utils.data.distributed.DistributedSampler(
                dataset, shuffle=is_train
            )

        else:
            if is_train:
                sampler = torch.utils.data.RandomSampler(dataset)
            else:
                sampler = torch.utils.data.SequentialSampler(dataset)
        return sampler

    def _create_loader(
        self,
        dataset,
        sampler,
        train,
        *,
        batch_size=1,
        num_workers=0,
        collate_fn=None,
    ):
        if self.config.system.distributed:
            if train:
                batch_sampler = torch.utils.data.BatchSampler(
                    sampler, batch_size, drop_last=False
                )

                loader = torch.utils.data.DataLoader(
                    dataset,
                    batch_sampler=batch_sampler,
                    num_workers=num_workers,
                    collate_fn=collate_fn,
                )
                return loader
            else:
                loader = torch.utils.data.DataLoader(
                    dataset,
                    batch_size=batch_size,
                    sampler=sampler,
                    num_workers=num_workers,
                    collate_fn=collate_fn,
                    drop_last=False,
                )
                return loader
        else:
            logger.info(f"not creating distributed dataloader")
            data_loader = torch.utils.data.DataLoader(
                dataset,
                batch_size=batch_size,
                sampler=sampler,
                num_workers=num_workers,
                drop_last=train,
                collate_fn=collate_fn,
            )
            return data_loader

    def _create_dryrun_dataset(self, dataset, subset_size):
        r = np.random.default_rng()
        indices = r.integers(0, len(dataset), subset_size)
        dataset = torch.utils.data.Subset(dataset, indices)
        return dataset

    def train(self, **kwargs):
        config = self.config
        train_dataset = Dataset.create(
            config.train.dataset.name,
            data_root=config.system.data_root,
            transforms=get_transform(),
            auth_token=config.system.auth_token,  # XXX (YC) Remove this
            **config.train.dataset.args,
        )
        logger.info(f"length of train dataset is {len(train_dataset)}")
        val_dataset = Dataset.create(
            config.val.dataset.name,
            data_root=config.system.data_root,
            transforms=get_transform(),
            auth_token=config.system.auth_token,  # XXX (YC) Remove this
            **config.val.dataset.args,
        )
        logger.info(f"length of validation dataset is {len(val_dataset)}")

        if self.config.system.dryrun:
            train_dataset = self._create_dryrun_dataset(
                dataset=train_dataset,
                subset_size=self.config.train.batch_size * 2,
            )
            val_dataset = self._create_dryrun_dataset(
                dataset=val_dataset, subset_size=self.config.val.batch_size * 2
            )
        train_sampler = self._create_sampler(
            dataset=train_dataset, is_train=True,
        )
        val_sampler = self._create_sampler(dataset=val_dataset, is_train=False)

        train_loader = self._create_loader(
            dataset=train_dataset,
            batch_size=self.config.train.batch_size,
            sampler=train_sampler,
            collate_fn=collate_fn,
            train=True,
        )
        val_loader = self._create_loader(
            dataset=val_dataset,
            batch_size=self.config.val.batch_size,
            sampler=val_sampler,
            collate_fn=collate_fn,
            train=False,
        )
        self.train_loop(
            config=self.config,
            train_dataloader=train_loader,
            label_mappings=train_dataset.label_mappings,
            val_dataloader=val_loader,
            writer=self.writer,
            model=self.model,
            train_sampler=train_sampler,
        )

    def save(self, path):
        """ Serialize Estimator to path

        Args:
            path (str): full path to save serialized estimator

        Returns:
            saved full path of the serialized estimator
        """
        save_dict = {
            "model": self.model_without_ddp.state_dict(),
            "config": self.config,
        }
        torch.save(save_dict, path)

        return path

    def load(self, path):
        """ Load Estimator from path

        Args:
            path (str): full path to the serialized estimator
        """
        logger.info(f"loading checkpoint from file")
        checkpoint = torch.load(path, map_location=self.device)
        self.model_without_ddp.load_state_dict(checkpoint["model"])
        loaded_config = copy.deepcopy(checkpoint["config"])
        stored_config = copy.deepcopy(self.config)
        del stored_config["checkpoint_file"]
        if stored_config != loaded_config:
            logger.warning(
                f"Found difference in estimator config."
                f"Estimator loaded from {path} was trained using config: "
                f"{loaded_config}. However, the current config is: "
                f"{self.config}."
            )

    def train_loop(
        self,
        *,
        model,
        config,
        writer,
        train_dataloader,
        label_mappings,
        val_dataloader,
        train_sampler=None,
    ):
        """

        Args:
            model: model to train
            config (CfgNode): estimator config
            writer: Tensorboard writer object
            device: model training on device (cpu|cuda)
            train_dataloader (torch.utils.data.DataLoader):
            label_mappings (dict): a dict of {label_id: label_name} mapping
            val_dataloader (torch.utils.data.DataLoader):
        """
        model = model.to(self.device)
        params = [p for p in model.parameters() if p.requires_grad]
        if config.optimizer.name == "Adam":
            optimizer = torch.optim.Adam(params, **config.optimizer.args)

            # use fixed learning rate when using Adam
            lr_scheduler = torch.optim.lr_scheduler.LambdaLR(
                optimizer, lambda x: 1.0
            )
        elif config.optimizer.name == "SGD":
            optimizer = torch.optim.SGD(
                params,
                lr=config.optimizer.args.lr,
                momentum=config.optimizer.args.momentum,
                weight_decay=float(config.optimizer.args.weight_decay),
            )
            lr_scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer,
                step_size=config.optimizer.args.lr_step_size,
                gamma=config.optimizer.args.lr_gamma,
            )
        else:
            raise ValueError(
                f"only valid optimizers are 'SGD' and 'Adam' but "
                f"received paramater {config.optimizer.name}"
            )

        accumulation_steps = config.train.get(
            "accumulation_steps", DEFAULT_ACCUMULATION_STEPS
        )
        logger.debug("Start training")
        total_timer = Timer(
            name="total-time", text=const.TIMING_TEXT, logger=logging.info
        )
        total_timer.start()
        for epoch in range(config.train.epochs):
            with Timer(
                name=f"epoch-{epoch}-train-time",
                text=const.TIMING_TEXT,
                logger=logging.info,
            ):
                train_one_epoch(
                    writer=writer,
                    model=model,
                    optimizer=optimizer,
                    data_loader=train_dataloader,
                    device=self.device,
                    epoch=epoch,
                    lr_scheduler=lr_scheduler,
                    accumulation_steps=accumulation_steps,
                    log_frequency=self.config.train.log_frequency,
                )
            if self.config.system.distributed:
                train_sampler.set_epoch(epoch)
            self.checkpointer.save(self, epoch=epoch)
            with Timer(
                name=f"epoch-{epoch}-evaluate-time",
                text=const.TIMING_TEXT,
                logger=logging.info,
            ):
                evaluate(
                    model=model,
                    data_loader=val_dataloader,
                    device=self.device,
                    writer=writer,
                    epoch=epoch,
                    metrics=self.metrics,
                    label_mappings=label_mappings,
                )
        total_timer.stop()



r"""Reference.

https://github.com/rafaelpadilla/Object-Detection-Metrics#average-precision\
Update algorithm from:
https://github.com/rafaelpadilla/Object-Detection-Metrics/blob/master/lib/Evaluator.py
"""
import collections

import numpy as np

from .base import EvaluationMetric
from .records import Records


class AveragePrecisionBBox2D(EvaluationMetric):
    """2D Bounding Box Average Precision metrics.

    Attributes:
        label_records (dict): save prediction records for each label
        ap_method (string): AP interoperation method name for AP calculation
        {"EveryPointInterpolation"| "NPointInterpolatedAP"}
        gt_bboxes_count (dict): ground truth box count for each label
        iou_threshold (float): iou threshold

    Args:
        iou_threshold (float): iou threshold (default: 0.5)
        interpolation (string): AP interoperation method name for AP calculation
    """

    def __init__(
        self, iou_threshold=0.5, interpolation="EveryPointInterpolation"
    ):
        if interpolation == "EveryPointInterpolation":
            self.ap_method = self.every_point_interpolated_ap
        elif interpolation == "NPointInterpolatedAP":
            self.ap_method = self.n_point_interpolated_ap
        else:
            raise ValueError(f"Unknown AP method name: {interpolation}!")

        self.iou_threshold = iou_threshold
        self.label_records = collections.defaultdict(
            lambda: Records(iou_threshold=self.iou_threshold)
        )
        self.gt_bboxes_count = collections.defaultdict(int)

    def reset(self):
        """Reset AP metrics."""
        self.label_records = collections.defaultdict(
            lambda: Records(iou_threshold=self.iou_threshold)
        )
        self.gt_bboxes_count = collections.defaultdict(int)

    def update(self, mini_batch):
        """Update records per mini batch

        Args:
            mini_batch (list(list)): a list which contains batch_size of
            gt bboxes and pred bboxes pair in each image.
            For example, if batch size = 2, mini_batch looks like:
            [[gt_bboxes1, pred_bboxes1], [gt_bboxes2, pred_bboxes2]]
            where gt_bboxes1, pred_bboxes1 contain gt bboxes and pred bboxes
            in one image
        """
        for bboxes in mini_batch:
            gt_bboxes, pred_bboxes = bboxes

            bboxes_per_label = self.label_bboxes(pred_bboxes)
            for label, boxes in bboxes_per_label.items():
                self.label_records[label].add_records(
                    gt_bboxes, boxes
                )

            for gt_bbox in gt_bboxes:
                self.gt_bboxes_count[gt_bbox.label] += 1

    def compute(self):
        """Compute AP for each label.

        Return:
            average_precision (dict): a dictionary of AP scores per label.
        """
        average_precision = {}
        label_records = self.label_records

        for label in self.gt_bboxes_count:
            # if there are no predicted boxes with this label
            if label not in label_records:
                average_precision[label] = 0
                continue
            pred_infos = label_records[label].pred_infos
            gt_bboxes_count = self.gt_bboxes_count[label]

            pred_infos = sorted(pred_infos, reverse=True)
            true_pos = np.array(list(zip(*pred_infos))[1]).astype(int)
            false_pos = 1 - true_pos

            acc_tp = np.cumsum(true_pos)
            acc_fp = np.cumsum(false_pos)

            recall = acc_tp / gt_bboxes_count
            precision = np.divide(acc_tp, (acc_fp + acc_tp))
            ap = self.ap_method(recall, precision)
            # add class result in the dictionary to be returned
            average_precision[label] = ap

        return average_precision

    @staticmethod
    def label_bboxes(pred_bboxes):
        """Save bboxes with same label in to a dictionary.

        Args:
            pred_bboxes (list): a list of prediction bounding boxes

        Returns:
            labels (dict): a dictionary of prediction bounding boxes
        """
        labels = collections.defaultdict(list)
        for box in pred_bboxes:
            labels[box.label].append(box)
        for label, boxes in labels.items():
            labels[label] = sorted(
                boxes, key=lambda bbox: bbox.score, reverse=True
            )
        return labels

    @staticmethod
    def every_point_interpolated_ap(recall, precision):
        """Calculating the interpolation performed in all points.

        Args:
            recall (list): recall history of the prediction
            precision (list): precision history of the prediction

        Returns: average precision for all points interpolation
        """
        # TODO: make it readable
        mrec = [0] + list(recall) + [1]
        mpre = [0] + list(precision) + [0]
        for i in range(len(mpre) - 1, 0, -1):
            mpre[i - 1] = max(mpre[i - 1], mpre[i])
        ii = []
        for i in range(len(mrec) - 1):
            if mrec[1:][i] != mrec[0:-1][i]:
                ii.append(i + 1)
        ap = 0
        for i in ii:
            ap = ap + np.sum((mrec[i] - mrec[i - 1]) * mpre[i])

        return ap

    @staticmethod
    def n_point_interpolated_ap(recall, precision, point=11):
        """Calculating the n-point interpolation.

        Args:
            recall (list): recall history of the prediction
            precision (list): precision history of the prediction
            point (int): n, n-point interpolation

        Returns: average precision for n-point interpolation
        """
        # TODO: make it readable
        mrec = [e for e in recall]
        mpre = [e for e in precision]
        recall_values = np.linspace(0, 1, point)
        recall_values = list(recall_values[::-1])
        rho_interp = []
        recall_valid = []
        # For each recall_values (0, 0.1, 0.2, ... , 1)
        for r in recall_values:
            # Obtain all recall values higher or equal than r
            arg_greater_recalls = np.argwhere(mrec[:] >= r)
            pmax = 0
            # If there are recalls above r
            if arg_greater_recalls.size != 0:
                pmax = max(mpre[arg_greater_recalls.min():])
            recall_valid.append(r)
            rho_interp.append(pmax)
        ap = sum(rho_interp) / point
        # Generating values for the plot
        rvals = []
        rvals.append(recall_valid[0])
        [rvals.append(e) for e in recall_valid]
        rvals.append(0)
        pvals = []
        pvals.append(0)
        [pvals.append(e) for e in rho_interp]
        pvals.append(0)

        cc = []
        for i in range(len(rvals)):
            p = (rvals[i], pvals[i - 1])
            if p not in cc:
                cc.append(p)
            p = (rvals[i], pvals[i])
            if p not in cc:
                cc.append(p)

        return ap



""" Save estimator checkpoints
"""
import logging
import os
import tempfile

import datasetinsights.constants as const
from datasetinsights.torch_distributed import is_master
from .gcs import GCSClient, gcs_bucket_and_path

logger = logging.getLogger(__name__)

DEFAULT_SUFFIX = "estimator"
GCS_BASE_STR = "gs://"


def checkpoint_file_on_gcs(config):
    """
    determine whether or not the checkpoint file to load an estimator is on
    gcs

    Args:
        config:

    Returns:

    """
    return (
        "checkpoint_file" in config.keys()
        and config.checkpoint_file != const.NULL_STRING
        and config.checkpoint_file[: len(GCS_BASE_STR)] == GCS_BASE_STR
    )


def create_checkpointer(*, logdir, config):
    """
    Initialize the correct estimator checkpointer
    Args:
        logdir: filepath to where to save/load local copy of estimator
        config:

    Returns (EstimatorCheckpoint): the correct estimator checkpoint for the
    config. The Estimator Checkpoint is responsible for saving and loading
    the estimator.

    """
    if logdir.startswith(GCS_BASE_STR):
        checkpointer = GCSEstimatorCheckpoint(logdir, config.estimator)
    elif checkpoint_file_on_gcs(config):
        logdir = f"{GCS_BASE_STR}{const.GCS_BUCKET}/runs/{str(logdir)}"
        checkpointer = GCSEstimatorCheckpoint(logdir, config.estimator)
    else:
        checkpointer = EstimatorCheckpoint(logdir, config.estimator)
    if config.system.distributed:
        checkpointer = DistributedEstimatorCheckpoint(
            is_master=is_master(), estimator_checkpoint=checkpointer
        )
    return checkpointer


class EstimatorCheckpoint:
    """ Interact with estimator checkpoints

    Args:
        dirname (str): directory name of where checkpoint files are stored
        prefix (str): filename prefix of the checkpoint files
        suffix (str): filename suffix of the checkpoint files
        create_dir (bool): indicate whether to force create directoy if
            the specified dirname does not exist. Default: True

    Attributes:
        dirname (str): directory name of where checkpoint files are stored
        prefix (str): filename prefix of the checkpoint files
        suffix (str): filename suffix of the checkpoint files
    """

    def __init__(
        self, dirname, prefix, *, suffix=DEFAULT_SUFFIX, create_dir=True
    ):
        self.dirname = dirname
        self.prefix = prefix
        self.suffix = suffix
        self.is_master = is_master()
        if create_dir:
            if not os.path.exists(dirname):
                os.makedirs(dirname)

        if not os.path.exists(dirname):
            raise ValueError(f"Directory path '{dirname}' is not found.")

    def save(self, estimator, epoch=None):
        """ Save estimator to checkpoint files.

        Args:
            estimator (datasetinsights.estimators.Estimator):
                datasetinsights estimator object.
            epoch (int): the current epoch number. Default: None

        Returns:
            full path to the saved checkpoint file
        """
        if epoch:
            filename = ".".join([self.prefix, f"ep{epoch}", self.suffix])
        else:
            filename = ".".join([self.prefix, self.suffix])
        path = os.path.join(self.dirname, filename)

        logger.debug(f"Saving estimator to {path}")
        estimator.save(path)

        return path

    def load(self, estimator, path):
        """ Load estimator from checkpoint files

        Args:
            estimator (datasetinsights.estimators.Estimator):
                datasetinsights estimator object.
            path: full path to the checkpoint file
        """
        logger.debug(f"Loading estimator from {path}")
        estimator.load(path)


class GCSEstimatorCheckpoint:
    """ Interact with estimator checkpoints on GCS

    Args:
        cloud_path (str): GCS cloud path (e.g. gs://bucket/path/to/directoy)
        prefix (str): filename prefix of the checkpoint files
        suffix (str): filename suffix of the checkpoint files
    """

    def __init__(self, cloud_path, prefix, *, suffix=DEFAULT_SUFFIX):
        self._tempdir = tempfile.TemporaryDirectory().name
        self._client = GCSClient()
        self._bucket, self._gcs_path = gcs_bucket_and_path(cloud_path)
        self._checkpointer = EstimatorCheckpoint(
            self._tempdir, prefix, create_dir=True, suffix=suffix
        )

    def save(self, estimator, epoch=None):
        """ Save estimator to checkpoint files on GCS

        Args:
            estimator (datasetinsights.estimators.Estimator):
                datasetinsights estimator object.
            epoch (int): the current epoch number. Default: None

        Returns:
            full GCS cloud path to the saved checkpoint file
        """
        path = self._checkpointer.save(estimator, epoch)
        filename = os.path.basename(path)
        object_key = os.path.join(self._gcs_path, filename)

        full_cloud_path = f"gs://{self._bucket}/{object_key}"

        logger.debug(f"Copying estimator from {path} to {full_cloud_path}")
        self._client.upload(path, self._bucket, object_key)

        return full_cloud_path

    def load(self, estimator, full_cloud_path):
        """ Load estimator from checkpoint files on GCS

        Args:
            estimator (datasetinsights.estimators.Estimator):
                datasetinsights estimator object.
            path: full path to the checkpoint file
        """
        bucket, object_key = gcs_bucket_and_path(full_cloud_path)
        filename = os.path.basename(object_key)
        path = os.path.join(self._tempdir, filename)
        logger.debug(f"Downloading estimator from {full_cloud_path} to {path}")
        self._client.download(bucket, object_key, path)

        self._checkpointer.load(estimator, path)


class DistributedEstimatorCheckpoint:
    def __init__(self, is_master, estimator_checkpoint):
        self.is_master = is_master
        self.estimator_checkpoint = estimator_checkpoint

    def save(self, estimator, epoch=None):
        if self.is_master:
            self.estimator_checkpoint.save(estimator, epoch)

    def load(self, estimator, full_cloud_path):
        if self.is_master:
            self.estimator_checkpoint.load(estimator, full_cloud_path)



""" Download Synthetic Dataset from Unity Simulation Platform

[Unity Simulation](https://unity.com/products/simulation) provides a powerful
platform for running simulations at large scale. This script provides basic
functionality that allow users to download generated synthetic dataset.
"""
import argparse
import logging
import os
from pathlib import Path

import datasetinsights.constants as const
from datasetinsights.data.simulation.download import (
    Downloader, download_manifest
)

logging.basicConfig(
    level=logging.INFO,
    format=(
        "%(levelname)s | %(asctime)s | %(name)s | %(threadName)s | "
        "%(message)s"
    ),
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(
        description="Download Dataset from Unity Simulation Platform"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="turn on verbose mode to enable debug logging",
    )
    parser.add_argument(
        "--data-root",
        type=str,
        default=const.DEFAULT_DATA_ROOT,
        help="root directory of datasets",
    )
    parser.add_argument(
        "--auth-token",
        type=str,
        help=(
            "authorization token to fetch USim manifest file that specified "
            "signed URLs for the simulation dataset files."
        ),
    )
    parser.add_argument(
        "--run-execution-id",
        type=str,
        help=(
            "USim run-execution-id"
        ),
        required=True,
    )
    parser.add_argument(
        "--include-binary",
        action="store_true",
        default=False,
        help="include binary files such as captured images.",
    )
    parser.add_argument(
        "--no-cache",
        action="store_true",
        default=False,
        help="toggle to force re-download by ignoring cache files",
    )

    args = parser.parse_args()

    return args


def run(args):
    if args.verbose:
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)

    logger.info("Run command with args: %s\n", args)
    data_root = args.data_root

    auth_token = args.auth_token
    use_cache = not args.no_cache
    run_execution_id = args.run_execution_id

    manifest_file = os.path.join(
        data_root, const.SYNTHETIC_SUBFOLDER, f"{run_execution_id}.csv"
    )
    if auth_token:
        manifest_file = download_manifest(
            run_execution_id,
            manifest_file,
            auth_token,
            use_cache=use_cache
        )
    else:
        logger.info(
            f"No auth token is provided. Assuming you already have "
            f"a manifest file located in {manifest_file}"
        )

    subfolder = Path(manifest_file).stem
    root = os.path.join(data_root, const.SYNTHETIC_SUBFOLDER, subfolder)

    dl_worker = Downloader(manifest_file, root, use_cache=use_cache)
    dl_worker.download_references()
    dl_worker.download_metrics()
    dl_worker.download_captures()
    if args.include_binary:
        dl_worker.download_binary_files()


if __name__ == "__main__":
    args = parse_args()
    run(args)



import concurrent.futures
import logging
import os
from pathlib import Path
import re

from codetiming import Timer
import numpy as np
import pandas as pd
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from tqdm import tqdm

from .exceptions import DownloadError
from .tables import DATASET_TABLES, FileType
import datasetinsights.constants as const


logger = logging.getLogger(__name__)
# number of workers for ThreadPoolExecutor. This is the default value
# in python3.8
MAX_WORKER = min(32, os.cpu_count() + 4)
# Timeout of requests (in seconds)
DEFAULT_TIMEOUT = 1800
# Retry after failed request
DEFAULT_MAX_RETRIES = 5


def _filter_unsuccessful_attempts(manifest_df):
    """
    remove all rows from a dataframe where a greater attempt_id exists for
    the 'instance_id'. This is necessary so that we avoid using data from
    a failed USim run and only get the most recent retry.
    Args:
        manifest_df (pandas df): must have columns 'attempt_id', 'app_param_id'
        and 'instance_id'

    Returns(pandas df): where all rows for earlier attempt ids have been
    removed

    """
    last_attempt_per_instance = manifest_df.groupby("instance_id")[
        "attempt_id"
    ].agg(["max"])
    merged = manifest_df.merge(
        how="outer",
        right=last_attempt_per_instance,
        left_on="instance_id",
        right_on="instance_id",
    )
    filtered = merged[merged["attempt_id"] == merged["max"]]
    filtered = filtered.reset_index(drop=True)
    filtered = filtered.drop(columns="max")
    return filtered


class Downloader:
    """Parse a given manifest file to download simulation output

    USim simulation docs:
    https://github.com/Unity-Technologies/Unity-Simulation-Docs

    Attributes:
        manifest (DataFrame): the csv manifest file stored in a pandas dataframe
        data_root (str): root directory where the simulation output should
            be downloaded
        use_cache (bool): use cache instead of re-download if file exists
    """

    MANIFEST_FILE_COLUMNS = (
        "run_execution_id",
        "app_param_id",
        "instance_id",
        "attempt_id",
        "file_name",
        "download_uri",
    )

    def __init__(
        self, manifest_file: str, data_root: str, use_cache: bool = True
    ):
        """ Initialize Downloader

        Args:
            manifest_file (str): path to a manifest file
            data_root (str): root directory where the simulation output should
                be downloaded
            use_cache (bool): use cache instead of re-download if file exists
        """
        self.manifest = pd.read_csv(
            manifest_file, header=0, names=self.MANIFEST_FILE_COLUMNS
        )
        self.manifest = _filter_unsuccessful_attempts(manifest_df=self.manifest)
        self.manifest["filetype"] = self.match_filetypes(self.manifest)
        self.data_root = data_root
        self.use_cache = use_cache

    @staticmethod
    def match_filetypes(manifest):
        """ Match filetypes for every rows in the manifest file.

        Args:
            manifest (pd.DataFrame): the manifest csv file

        Returns:
            a list of filetype strings
        """
        filenames = manifest.file_name
        filetypes = []
        for name in filenames:
            for _, table in DATASET_TABLES.items():
                if re.match(table.pattern, name):
                    filetypes.append(table.filetype)
                    break
            else:
                filetypes.append(FileType.BINARY)

        return filetypes

    @Timer(name="download_all", text=const.TIMING_TEXT, logger=logging.info)
    def download_all(self):
        """ Download all files in the manifest file.
        """
        matched_rows = np.ones(len(self.manifest), dtype=bool)
        downloaded = self._download_rows(matched_rows)
        logger.info(
            f"Total {len(downloaded)} files in manifest are successfully "
            f"downloaded."
        )

    @Timer(
        name="download_references", text=const.TIMING_TEXT, logger=logging.info
    )
    def download_references(self):
        """ Download all reference files.
        #todo update link
        https://docs.google.com/document/d/1lKPm06z09uX9gZIbmBUMO6WKlIGXiv3hgXb_taPOnU0/edit#bookmark=id.sjklr8srzpt8
        """
        logger.info("Downloading references files...")
        matched_rows = self.manifest.filetype == FileType.REFERENCE
        downloaded = self._download_rows(matched_rows)

        logger.info(
            f"Total {len(downloaded)} reference files are successfully "
            f"downloaded."
        )

    @Timer(name="download_metrics", text=const.TIMING_TEXT, logger=logging.info)
    def download_metrics(self):
        """ Download all metrics files.
        #todo update link
        https://docs.google.com/document/d/1lKPm06z09uX9gZIbmBUMO6WKlIGXiv3hgXb_taPOnU0/edit#bookmark=id.62bm95fh1g2h
        """
        logger.info("Downloading metrics files...")
        matched_rows = self.manifest.filetype == FileType.METRIC
        downloaded = self._download_rows(matched_rows)
        logger.info(
            f"Total {len(downloaded)} metric files are successfully downloaded."
        )

    @Timer(
        name="download_captures", text=const.TIMING_TEXT, logger=logging.info
    )
    def download_captures(self):
        """ Download all captures files.
        #todo update links
        https://docs.google.com/document/d/1lKPm06z09uX9gZIbmBUMO6WKlIGXiv3hgXb_taPOnU0/edit#bookmark=id.ueuu07ww8otx
        """
        logger.info("Downloading captures files...")
        matched_rows = self.manifest.filetype == FileType.CAPTURE
        downloaded = self._download_rows(matched_rows)
        logger.info(
            f"Total {len(downloaded)} capture files are successfully "
            f"downloaded."
        )

    @Timer(
        name="download_binary_files",
        text=const.TIMING_TEXT,
        logger=logging.info,
    )
    def download_binary_files(self):
        """ Download all binary files.
        """
        logger.info("Downloading binary files...")
        matched_rows = self.manifest.filetype == FileType.BINARY
        downloaded = self._download_rows(matched_rows)
        logger.info(
            f"Total {len(downloaded)} binary files are successfully "
            f"downloaded."
        )

    def _download_rows(self, matched_rows):
        """ Download matched rows in a manifest file.

        Note:
        We might need to download 1M+ of simulation output files, in this case
        we don't want to have a single file transfer failure holding back on
        getting the simulation data. Here download exception are captured.
        We only log an error message and requires uses to pay attention to
        this error.

        Args:
            matched_rows (pd.Series): boolean series indicator of the manifest
                file that should be downloaded

        Returns:
            list of strings representing the downloaded destination path.
        """
        n_expected = sum(matched_rows)
        future_downloaded = []
        downloaded = []
        with concurrent.futures.ThreadPoolExecutor(MAX_WORKER) as executor:
            for _, row in self.manifest[matched_rows].iterrows():
                source_uri = row.download_uri
                relative_path = row.file_name
                dest_path = os.path.join(self.data_root, relative_path)
                future = executor.submit(
                    _download_file, source_uri, dest_path, self.use_cache
                )
                future_downloaded.append(future)

            for future in tqdm(
                concurrent.futures.as_completed(future_downloaded),
                total=n_expected,
            ):
                try:
                    downloaded.append(future.result())
                except DownloadError as ex:
                    logger.error(ex)

        n_downloaded = len(downloaded)
        if n_downloaded != n_expected:
            logger.warning(
                f"Found {n_expected} matching records in the manifest file, "
                f"but only {n_downloaded} are downloaded."
            )

        return downloaded


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, timeout, *args, **kwargs):
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def _download_file(source_uri: str, dest_path: str, use_cache: bool = True):
    """Download a file specified from a source uri

    Args:
        source_uri (str): source url where the file should be downloaded
        dest_path (str): destination path of the file
        use_cache (bool): use_cache (bool): use cache instead of
                re-download if file exists

    Returns:
        String of destination path.
    """
    dest_path = Path(dest_path)
    if dest_path.exists() and use_cache:
        return dest_path

    logger.debug(f"Trying to download file from {source_uri} -> {dest_path}")
    adapter = TimeoutHTTPAdapter(
        timeout=DEFAULT_TIMEOUT,
        max_retries=Retry(total=DEFAULT_MAX_RETRIES)
    )
    with requests.Session() as http:
        http.mount("https://", adapter)
        try:
            response = http.get(source_uri)
            response.raise_for_status()
        except requests.exceptions.RequestException as ex:
            logger.error(ex)
            err_msg = (
                f"The request download from {source_uri} -> {dest_path} can't "
                f"be completed."
            )

            raise DownloadError(err_msg)
        else:
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            with open(dest_path, "wb") as f:
                f.write(response.content)

    return dest_path


def download_manifest(
    run_execution_id,
    manifest_file,
    auth_token,
    project_id,
    use_cache=True
):
    """ Download manifest file from a single run_execution_id
    #todo are these docs public?
    https://github.com/Unity-Technologies/Unity-Simulation-Docs
    Args:
        run_execution_id (str): Unity Simulation run execution id
        manifest_file (str): path to the destination of the manifest_file
        auth_token (str): short lived authorization token
        project_id (str): Unity project id that has Unity Simulation enabled
        use_cache (bool, optional): indicator to skip download if manifest
            file already exists. Default: True.

    Returns:
        str: Full path to the manifest_file
    """
    api_endpoint = const.USIM_API_ENDPOINT
    project_url = f"{api_endpoint}/v1/projects/{project_id}/"
    data_url = f"{project_url}runs/{run_execution_id}/data"
    if Path(manifest_file).exists() and use_cache:
        logger.info(
            f"Mainfest file {manifest_file} already exists. Skipping downloads."
        )
        return manifest_file

    logger.info(
        f"Trying to download manifest file for run-execution-id "
        f"{run_execution_id}"
    )
    adapter = TimeoutHTTPAdapter(
        timeout=DEFAULT_TIMEOUT,
        max_retries=Retry(total=DEFAULT_MAX_RETRIES)
    )
    headers = {
        "Authorization": f"Bearer {auth_token}",
        "Content-Type": "application/json",
    }
    with requests.Session() as http:
        http.mount("https://", adapter)
        try:
            resp = http.get(data_url, headers=headers)
            resp.raise_for_status()
        except requests.exceptions.RequestException as ex:
            logger.error(ex)
            err_msg = (
                f"Failed to download manifest file for run-execution-id: "
                f"{run_execution_id}."
            )
            raise DownloadError(err_msg)
        else:
            Path(manifest_file).parent.mkdir(parents=True, exist_ok=True)
            with open(manifest_file, "wb") as f:
                for chunk in resp.iter_content(chunk_size=1024):
                    f.write(chunk)

    logger.info(
        f"Manifest file {manifest_file} downloaded for run-execution-id "
        f"{run_execution_id}"
    )

    return manifest_file



import argparse
import logging
import sys

from tensorboardX import SummaryWriter
import torch
from yacs.config import CfgNode as CN

import datasetinsights.constants as const
from .configs import system
from .data.datasets import Dataset
from .estimators import Estimator
from .storage.checkpoint import create_checkpointer
from .torch_distributed import init_distributed_mode, get_world_size, is_master

logging.basicConfig(
    level=logging.INFO,
    format=(
        "%(levelname)s | %(asctime)s | %(name)s | %(threadName)s | "
        "%(message)s"
    ),
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger(__name__)


def parse_args():
    command_parser = argparse.ArgumentParser(
        description="Datasetinsights Modeling Interface"
    )
    command_parser.add_argument(
        "command",
        help="sub-command to run",
        choices=("train", "evaluate", "download-train", "download-evaluate"),
    )

    parser = argparse.ArgumentParser(
        description="Datasetinsights Modeling Interface"
    )
    parser.add_argument(
        "--local_rank",
        help="the local rank of the subprocess excecuting the code. "
        "If sequential, this should be 0."
        "Otherwise torch.distributed.launch will specify this arg.",
        type=int,
    )
    parser.add_argument(
        "--gpus-per-node",
        action="store_true",
        default=get_world_size(),
        help="number of gpus per node to use for training/evaluating",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=system.verbose,
        help="turn on verbose mode to enable debug logging",
    )
    parser.add_argument(
        "--dryrun",
        action="store_true",
        default=system.dryrun,
        help="dry run with only a small subset of the dataset",
    )
    parser.add_argument(
        "--no-cuda",
        action="store_true",
        default=system.no_cuda,
        help="force to turn off cuda training",
    )
    parser.add_argument(
        "--logdir",
        type=str,
        default=system.logdir,
        help="path where to save training logs",
    )
    parser.add_argument(
        "--auth-token",
        type=str,
        default=system.auth_token,
        help=(
            "authorization token to fetch USim manifest file that speficed "
            "signed URLs for the simulation dataset files."
        ),
    )
    parser.add_argument(
        "--data-root",
        type=str,
        default=system.data_root,
        help="root directory of all datasets",
    )
    parser.add_argument(
        "--val-interval",
        type=int,
        default=system.val_interval,
        help="control how many training epochs between each validation run",
    )
    parser.add_argument(
        "-j",
        "--workers",
        type=int,
        metavar="N",
        default=system.workers,
        help="number of data loading workers",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        required=True,
        help="config file for this model",
    )
    parser.add_argument(
        "opts",
        default=None,
        nargs=argparse.REMAINDER,
        help="extra options to override config file",
    )
    cmd = command_parser.parse_args(sys.argv[2:3]).command
    cmd_args = parser.parse_args([sys.argv[1]] + sys.argv[3:])
    logger.debug("Parsed CLI args: %s", cmd_args)

    return cmd, cmd_args


def parse_config(args):
    cfg = CN.load_cfg(open(args.config, "r"))
    cfg.merge_from_list(args.opts)

    system.logdir = args.logdir
    system.data_root = args.data_root
    system.verbose = args.verbose
    system.dryrun = args.dryrun
    system.no_cuda = args.no_cuda
    system.workers = args.workers
    system.val_interval = args.val_interval
    system.auth_token = args.auth_token
    system.distributed = args.distributed
    system.gpus_per_node = args.gpus_per_node
    cfg.system = system
    cfg.freeze()
    return cfg


def run(command, cfg):
    if cfg.system.verbose:
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)

    logger.info("Run command: %s with config: %s\n", command, cfg)

    if torch.cuda.is_available() and not cfg.system.no_cuda:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")
    logdir = cfg.system.logdir
    if logdir == const.NULL_STRING:
        # Use logdir=None to force using SummaryWriter default logdir,
        # which points to ./runs/<model>_<timestamp>
        logdir = None

    # todo this makes it so that we lose the tensorboard writer of non-master
    # processes which could make debugging harder
    writer = SummaryWriter(logdir, write_to_disk=is_master())
    checkpointer = create_checkpointer(logdir=writer.logdir, config=cfg)
    estimator = Estimator.create(
        cfg.estimator,
        config=cfg,
        writer=writer,
        device=device,
        checkpointer=checkpointer,
        gpu=args.gpu,
        rank=args.rank,
    )

    if command == "train":
        estimator.train()
    elif command == "evaluate":
        estimator.evaluate()
    elif command == "download-train":
        # TODO (YC)
        # We should remove reference to auth-token in various places to
        # enable download synthetic dataset. Usim is working on a solution
        # that will enable customers to sprcify cloud storage path
        # to store simulations. In the future, we should simply rely
        # on gcs service accounts to access simulation data for a given
        # run execution id.
        Dataset.create(
            cfg.train.dataset.name,
            data_root=cfg.system.data_root,
            auth_token=cfg.system.auth_token,  # XXX(YC) This should be removed
            **cfg.train.dataset.args,
        )
        Dataset.create(
            cfg.val.dataset.name,
            data_root=cfg.system.data_root,
            auth_token=cfg.system.auth_token,  # XXX(YC) This should be removed
            **cfg.val.dataset.args,
        )
    elif command == "download-evaluate":
        Dataset.create(
            cfg.test.dataset.name,
            data_root=cfg.system.data_root,
            auth_token=cfg.system.auth_token,  # XXX(YC) This should be removed
            **cfg.test.dataset.args,
        )

    writer.close()


if __name__ == "__main__":
    command, args = parse_args()
    init_distributed_mode(args)
    cfg = parse_config(args)
    run(command, cfg)



"""System configurations
"""
from yacs.config import CfgNode as CN

import datasetinsights.constants as const

system = CN()

system.logdir = const.NULL_STRING
system.data_root = const.DEFAULT_DATA_ROOT
system.verbose = False
system.dryrun = False
system.no_cuda = False
system.workers = 0
system.val_interval = 1

# USim authorization token
system.auth_token = const.NULL_STRING



from .captures import Captures
from .references import AnnotationDefinitions, MetricDefinitions, Egos, Sensors
from .metrics import Metrics


__all__ = [
    AnnotationDefinitions,
    Captures,
    Egos,
    Metrics,
    MetricDefinitions,
    Sensors,
]



GCS_BUCKET = "thea-dev"

# This is a hack on yacs config system, as it does not allow null values
# in configs. They are working on supporting null values in config
# https://github.com/rbgirshick/yacs/pull/18.
NULL_STRING = "None"

# Root directory of all datasets
# We assume the datasets are stored in the following structure:
# data_root/
#   cityscapes/
#   kitti/
#   nuscenes/
#   synthetic/
#   ...
DEFAULT_DATA_ROOT = "/data"
SYNTHETIC_SUBFOLDER = "synthetic"

# Default Unity Project ID where USim jobs was executed
DEFAULT_PROJECT_ID = "474ba200-4dcc-4976-818e-0efd28efed30"
USIM_API_ENDPOINT = "https://api.simulation.unity3d.com"

# Default Timing text for codetiming.Timer decorator
TIMING_TEXT = "[{name}] elapsed time: {:0.4f} seconds."



from .average_log10_error import AverageLog10Error
from .average_precision_2d_bbox import AveragePrecisionBBox2D
from .average_recall_2d_bbox import AverageRecallBBox2D
from .average_relative_error import AverageRelativeError
from .base import EvaluationMetric
from .iou import IoU
from .root_mean_square_error import RootMeanSquareError
from .threshold_accuracy import ThresholdAccuracy


__all__ = [
    EvaluationMetric,
    IoU,
    AverageRelativeError,
    AverageLog10Error,
    AveragePrecisionBBox2D,
    AverageRecallBBox2D,
    RootMeanSquareError,
    ThresholdAccuracy,
]
